@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 """
 notation3.py - Standalone Notation3 Parser
 Derived from CWM, the Closed World Machine
@@ -27,19 +26,49 @@ Modified to work with rdflib by Gunnar Aastrand Grimnes
 Copyright 2010, Gunnar A. Grimnes
 
 """
-import sys
+
+from __future__ import annotations
+
+import codecs
 import os
 import re
-import codecs
+import sys
 
+# importing typing for `typing.List` because `List`` is used for something else
+import typing
 from decimal import Decimal
-
+from typing import (
+    IO,
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Match,
+    MutableSequence,
+    NoReturn,
+    Optional,
+    Pattern,
+    Set,
+    Tuple,
+    TypeVar,
+    Union,
+)
 from uuid import uuid4
 
-from rdflib.exceptions import ParserError
-from rdflib.term import URIRef, BNode, Literal, Variable, _XSD_PFX, _unique_id
-from rdflib.graph import QuotedGraph, ConjunctiveGraph, Graph
 from rdflib.compat import long_type
+from rdflib.exceptions import ParserError
+from rdflib.graph import Dataset, Graph, QuotedGraph
+from rdflib.term import (
+    _XSD_PFX,
+    BNode,
+    IdentifiedNode,
+    Identifier,
+    Literal,
+    Node,
+    URIRef,
+    Variable,
+    _unique_id,
+)
 
 __all__ = [
     "BadSyntax",
@@ -51,12 +80,20 @@ __all__ = [
     "runNamespace",
     "uniqueURI",
     "hexify",
+    "Formula",
+    "RDFSink",
+    "SinkParser",
 ]
 
 from rdflib.parser import Parser
 
+if TYPE_CHECKING:
+    from rdflib.parser import InputSource
 
-def splitFragP(uriref, punct=0):
+_AnyT = TypeVar("_AnyT")
+
+
+def splitFragP(uriref: str, punc: int = 0) -> Tuple[str, str]:
     """split a URI reference before the fragment
 
     Punctuation is kept.
@@ -78,7 +115,10 @@ def splitFragP(uriref, punct=0):
         return uriref, ""
 
 
-def join(here, there):
+_StrT = TypeVar("_StrT", bound=str)
+
+
+def join(here: str, there: str) -> str:
     """join an absolute URI and URI reference
     (non-ascii characters are supported/doctested;
     haven't checked the details of the IRI spec though)
@@ -106,11 +146,11 @@ def join(here, there):
 
     We grok IRIs
 
-    >>> len(u'Andr\\xe9')
+    >>> len('Andr\\xe9')
     5
 
-    >>> join('http://example.org/', u'#Andr\\xe9')
-    u'http://example.org/#Andr\\xe9'
+    >>> join('http://example.org/', '#Andr\\xe9')
+    'http://example.org/#Andr\\xe9'
     """
 
     #    assert(here.find("#") < 0), \
@@ -133,7 +173,7 @@ def join(here, there):
         return here + frag
 
     # join('mid:foo@example', '../foo') bzzt
-    if here[bcolonl + 1] != "/":
+    if here[bcolonl + 1 : bcolonl + 2] != "/":
         raise ValueError(
             "Base <%s> has no slash after "
             "colon - with relative '%s'." % (here, there)
@@ -176,7 +216,7 @@ def join(here, there):
     return here[: slashr + 1] + path + frag
 
 
-def base():
+def base() -> str:
     """The base URI for this process - the Web equiv of cwd
 
     Relative or absolute unix-standard filenames parsed relative to
@@ -189,7 +229,7 @@ def base():
     return "file://" + _fixslash(os.getcwd()) + "/"
 
 
-def _fixslash(s):
+def _fixslash(s: str) -> str:
     """Fix windowslike filename to unixlike - (#ifdef WINDOWS)"""
     s = s.replace("\\", "/")
     if s[0] != "/" and s[1] == ":":
@@ -236,10 +276,10 @@ N3_List = (SYMBOL, List_NS + "List")
 N3_Empty = (SYMBOL, List_NS + "Empty")
 
 
-runNamespaceValue = None
+runNamespaceValue: Optional[str] = None
 
 
-def runNamespace():
+def runNamespace() -> str:
     """Returns a URI suitable as a namespace for run-local objects"""
     # @@@ include hostname (privacy?) (hash it?)
     global runNamespaceValue
@@ -251,7 +291,7 @@ def runNamespace():
 nextu = 0
 
 
-def uniqueURI():
+def uniqueURI() -> str:
     """A unique URI"""
     global nextu
     nextu += 1
@@ -264,12 +304,12 @@ chatty_flag = 50
 # from why import BecauseOfData, becauseSubexpression
 
 
-def BecauseOfData(*args, **kargs):
+def BecauseOfData(*args: Any, **kargs: Any) -> None:
     # print args, kargs
     pass
 
 
-def becauseSubexpression(*args, **kargs):
+def becauseSubexpression(*args: Any, **kargs: Any) -> None:
     # print args, kargs
     pass
 
@@ -310,10 +350,10 @@ numberChars = set("0123456789-")
 numberCharsPlus = numberChars | {"+", "."}
 
 
-def unicodeExpand(m):
+def unicodeExpand(m: Match) -> str:
     try:
         return chr(int(m.group(1), 16))
-    except:
+    except Exception:
         raise Exception("Invalid unicode code point: " + m.group(1))
 
 
@@ -343,13 +383,13 @@ langcode = re.compile(r"[a-zA-Z0-9]+(-[a-zA-Z0-9]+)*")
 class SinkParser:
     def __init__(
         self,
-        store,
-        openFormula=None,
-        thisDoc="",
-        baseURI=None,
-        genPrefix="",
-        why=None,
-        turtle=False,
+        store: RDFSink,
+        openFormula: Optional[Formula] = None,
+        thisDoc: str = "",
+        baseURI: Optional[str] = None,
+        genPrefix: str = "",
+        why: Optional[Callable[[], None]] = None,
+        turtle: bool = False,
     ):
         """note: namespace names should *not* end in  # ;
         the  # will get added during qname processing"""
@@ -361,7 +401,8 @@ class SinkParser:
 
         self._store = store
         if genPrefix:
-            store.setGenPrefix(genPrefix)  # pass it on
+            # TODO FIXME: there is no function named setGenPrefix
+            store.setGenPrefix(genPrefix)  # type: ignore[attr-defined] # pass it on
 
         self._thisDoc = thisDoc
         self.lines = 0  # for error handling
@@ -369,10 +410,10 @@ class SinkParser:
         self._genPrefix = genPrefix
         self.keywords = ["a", "this", "bind", "has", "is", "of", "true", "false"]
         self.keywordsSet = 0  # Then only can others be considered qnames
-        self._anonymousNodes = {}
+        self._anonymousNodes: Dict[str, BNode] = {}
         # Dict of anon nodes already declared ln: Term
-        self._variables = {}
-        self._parentVariables = {}
+        self._variables: Dict[str, Variable] = {}
+        self._parentVariables: Dict[str, Variable] = {}
         self._reason = why  # Why the parser was asked to parse this
 
         self.turtle = turtle  # raise exception when encountering N3 extensions
@@ -380,13 +421,15 @@ class SinkParser:
         # only allows double quotes.
         self.string_delimiters = ('"', "'") if turtle else ('"',)
 
-        self._reason2 = None  # Why these triples
+        self._reason2: Optional[Callable[..., None]] = None  # Why these triples
         # was: diag.tracking
         if tracking:
-            self._reason2 = BecauseOfData(
+            # type error: "BecauseOfData" does not return a value
+            self._reason2 = BecauseOfData(  # type: ignore[func-returns-value]
                 store.newSymbol(thisDoc), because=self._reason
             )
 
+        self._baseURI: Optional[str]
         if baseURI:
             self._baseURI = baseURI
         else:
@@ -403,18 +446,20 @@ class SinkParser:
             else:
                 self._genPrefix = uniqueURI()
 
+        self._formula: Optional[Formula]
         if openFormula is None and not turtle:
             if self._thisDoc:
-                self._formula = store.newFormula(thisDoc + "#_formula")
+                # TODO FIXME: store.newFormula does not take any arguments
+                self._formula = store.newFormula(thisDoc + "#_formula")  # type: ignore[call-arg]
             else:
                 self._formula = store.newFormula()
         else:
             self._formula = openFormula
 
-        self._context = self._formula
-        self._parentContext = None
+        self._context: Optional[Formula] = self._formula
+        self._parentContext: Optional[Formula] = None
 
-    def here(self, i):
+    def here(self, i: int) -> str:
         """String generated from position in file
 
         This is for repeatability when referring people to bnodes in a document.
@@ -427,20 +472,20 @@ class SinkParser:
 
         return "%s_L%iC%i" % (self._genPrefix, self.lines, i - self.startOfLine + 1)
 
-    def formula(self):
+    def formula(self) -> Optional[Formula]:
         return self._formula
 
-    def loadStream(self, stream):
+    def loadStream(self, stream: Union[IO[str], IO[bytes]]) -> Optional[Formula]:
         return self.loadBuf(stream.read())  # Not ideal
 
-    def loadBuf(self, buf):
+    def loadBuf(self, buf: Union[str, bytes]) -> Optional[Formula]:
         """Parses a buffer and returns its top level formula"""
         self.startDoc()
 
         self.feed(buf)
         return self.endDoc()  # self._formula
 
-    def feed(self, octets):
+    def feed(self, octets: Union[str, bytes]) -> None:
         """Feed an octet stream to the parser
 
         if BadSyntax is raised, the string
@@ -468,8 +513,7 @@ class SinkParser:
                 # print("# next char: %s" % s[j])
                 self.BadSyntax(s, j, "expected directive or statement")
 
-    def directiveOrStatement(self, argstr, h):
-
+    def directiveOrStatement(self, argstr: str, h: int) -> int:
         i = self.skipSpace(argstr, h)
         if i < 0:
             return i  # EOF
@@ -492,7 +536,7 @@ class SinkParser:
     # @@I18N
     # _namechars = string.lowercase + string.uppercase + string.digits + '_-'
 
-    def tok(self, tok, argstr, i, colon=False):
+    def tok(self, tok: str, argstr: str, i: int, colon: bool = False) -> int:
         """Check for keyword.  Space must have been stripped on entry and
         we must not be at end of file.
 
@@ -517,7 +561,7 @@ class SinkParser:
         else:
             return -1
 
-    def sparqlTok(self, tok, argstr, i):
+    def sparqlTok(self, tok: str, argstr: str, i: int) -> int:
         """Check for SPARQL keyword.  Space must have been stripped on entry
         and we must not be at end of file.
         Case insensitive and not preceded by @
@@ -534,11 +578,11 @@ class SinkParser:
         else:
             return -1
 
-    def directive(self, argstr, i):
+    def directive(self, argstr: str, i: int) -> int:
         j = self.skipSpace(argstr, i)
         if j < 0:
             return j  # eof
-        res = []
+        res: typing.List[str] = []
 
         j = self.tok("bind", argstr, i)  # implied "#". Obsolete.
         if j > 0:
@@ -568,7 +612,8 @@ class SinkParser:
             for x in res:
                 # self._context.declareUniversal(x)
                 if x not in self._variables or x in self._parentVariables:
-                    self._variables[x] = self._context.newUniversal(x)
+                    # type error: Item "None" of "Optional[Formula]" has no attribute "newUniversal"
+                    self._variables[x] = self._context.newUniversal(x)  # type: ignore[union-attr]
             return i
 
         j = self.tok("forSome", argstr, i)
@@ -580,19 +625,20 @@ class SinkParser:
             if i < 0:
                 self.BadSyntax(argstr, i, "Bad variable list after @forSome")
             for x in res:
-                self._context.declareExistential(x)
+                # type error: Item "None" of "Optional[Formula]" has no attribute "declareExistential"
+                self._context.declareExistential(x)  # type: ignore[union-attr]
             return i
 
         j = self.tok("prefix", argstr, i, colon=True)  # no implied "#"
         if j >= 0:
-            t = []
+            t: typing.List[Union[Identifier, Tuple[str, str]]] = []
             i = self.qname(argstr, j, t)
             if i < 0:
                 self.BadSyntax(argstr, j, "expected qname after @prefix")
             j = self.uri_ref2(argstr, i, t)
             if j < 0:
                 self.BadSyntax(argstr, i, "expected <uriref> after @prefix _qname_")
-            ns = self.uriOf(t[1])
+            ns: str = self.uriOf(t[1])
 
             if self._baseURI:
                 ns = join(self._baseURI, ns)
@@ -632,8 +678,7 @@ class SinkParser:
 
         return -1  # Not a directive, could be something else.
 
-    def sparqlDirective(self, argstr, i):
-
+    def sparqlDirective(self, argstr: str, i: int) -> int:
         """
         turtle and trig support BASE/PREFIX without @ and without
         terminating .
@@ -645,7 +690,7 @@ class SinkParser:
 
         j = self.sparqlTok("PREFIX", argstr, i)
         if j >= 0:
-            t = []
+            t: typing.List[Any] = []
             i = self.qname(argstr, j, t)
             if i < 0:
                 self.BadSyntax(argstr, j, "expected qname after @prefix")
@@ -695,14 +740,14 @@ class SinkParser:
 
         return -1  # Not a directive, could be something else.
 
-    def bind(self, qn, uri):
+    def bind(self, qn: str, uri: bytes) -> None:
         assert isinstance(uri, bytes), "Any unicode must be %x-encoded already"
         if qn == "":
             self._store.setDefaultNamespace(uri)
         else:
             self._store.bind(qn, uri)
 
-    def setKeywords(self, k):
+    def setKeywords(self, k: Optional[typing.List[str]]) -> None:
         """Takes a list of strings"""
         if k is None:
             self.keywordsSet = 0
@@ -710,22 +755,22 @@ class SinkParser:
             self.keywords = k
             self.keywordsSet = 1
 
-    def startDoc(self):
+    def startDoc(self) -> None:
         # was: self._store.startDoc()
         self._store.startDoc(self._formula)
 
-    def endDoc(self):
+    def endDoc(self) -> Optional[Formula]:
         """Signal end of document and stop parsing. returns formula"""
         self._store.endDoc(self._formula)  # don't canonicalize yet
         return self._formula
 
-    def makeStatement(self, quadruple):
+    def makeStatement(self, quadruple) -> None:
         # $$$$$$$$$$$$$$$$$$$$$
         # print "# Parser output: ", `quadruple`
         self._store.makeStatement(quadruple, why=self._reason2)
 
-    def statement(self, argstr, i):
-        r = []
+    def statement(self, argstr: str, i: int) -> int:
+        r: typing.List[Any] = []
         i = self.object(argstr, i, r)  # Allow literal for subject - extends RDF
         if i < 0:
             return i
@@ -736,10 +781,10 @@ class SinkParser:
             self.BadSyntax(argstr, i, "expected propertylist")
         return j
 
-    def subject(self, argstr, i, res):
+    def subject(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         return self.item(argstr, i, res)
 
-    def verb(self, argstr, i, res):
+    def verb(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         """has _prop_
         is _prop_ of
         a
@@ -753,7 +798,7 @@ class SinkParser:
         if j < 0:
             return j  # eof
 
-        r = []
+        r: typing.List[Any] = []
 
         j = self.tok("has", argstr, i)
         if j >= 0:
@@ -825,16 +870,16 @@ class SinkParser:
 
         return -1
 
-    def prop(self, argstr, i, res):
+    def prop(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         return self.item(argstr, i, res)
 
-    def item(self, argstr, i, res):
+    def item(self, argstr: str, i, res: MutableSequence[Any]) -> int:
         return self.path(argstr, i, res)
 
-    def blankNode(self, uri=None):
+    def blankNode(self, uri: Optional[str] = None) -> BNode:
         return self._store.newBlankNode(self._context, uri, why=self._reason2)
 
-    def path(self, argstr, i, res):
+    def path(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         """Parse the path production."""
         j = self.nodeOrLiteral(argstr, i, res)
         if j < 0:
@@ -855,7 +900,7 @@ class SinkParser:
             res.append(obj)
         return j
 
-    def anonymousNode(self, ln):
+    def anonymousNode(self, ln: str) -> BNode:
         """Remember or generate a term for one of these _: anonymous nodes"""
         term = self._anonymousNodes.get(ln, None)
         if term is not None:
@@ -864,12 +909,18 @@ class SinkParser:
         self._anonymousNodes[ln] = term
         return term
 
-    def node(self, argstr, i, res, subjectAlready=None):
+    def node(
+        self,
+        argstr: str,
+        i: int,
+        res: MutableSequence[Any],
+        subjectAlready: Optional[Node] = None,
+    ) -> int:
         """Parse the <node> production.
         Space is now skipped once at the beginning
         instead of in multiple calls to self.skipSpace().
         """
-        subj = subjectAlready
+        subj: Optional[Node] = subjectAlready
 
         j = self.skipSpace(argstr, i)
         if j < 0:
@@ -889,7 +940,7 @@ class SinkParser:
                         argstr, j, "Found '[=' or '[ =' when in turtle mode."
                     )
                 i = j + 1
-                objs = []
+                objs: typing.List[Node] = []
                 j = self.objectList(argstr, i, objs)
                 if j >= 0:
                     subj = objs[0]
@@ -950,7 +1001,7 @@ class SinkParser:
                     else:
                         first_run = False
 
-                    item = []
+                    item: typing.List[Any] = []
                     j = self.item(argstr, i, item)  # @@@@@ should be path, was object
                     if j < 0:
                         self.BadSyntax(argstr, i, "expected item in set or '$}'")
@@ -970,8 +1021,10 @@ class SinkParser:
                 reason2 = self._reason2
                 self._reason2 = becauseSubexpression
                 if subj is None:
-                    subj = self._store.newFormula()
-                self._context = subj
+                    # type error: Incompatible types in assignment (expression has type "Formula", variable has type "Optional[Node]")
+                    subj = self._store.newFormula()  # type: ignore[assignment]
+                # type error: Incompatible types in assignment (expression has type "Optional[Node]", variable has type "Optional[Formula]")
+                self._context = subj  # type: ignore[assignment]
 
                 while 1:
                     i = self.skipSpace(argstr, j)
@@ -992,10 +1045,16 @@ class SinkParser:
                 self._context = self._parentContext
                 self._reason2 = reason2
                 self._parentContext = oldParentContext
-                res.append(subj.close())  # No use until closed
+                # type error: Item "Node" of "Optional[Node]" has no attribute "close"
+                res.append(
+                    subj.close()  # type: ignore[union-attr]
+                )  # No use until closed
                 return j
 
         if ch == "(":
+            thing_type: Callable[
+                [typing.List[Any], Optional[Formula]], Union[Set[Any], IdentifiedNode]
+            ]
             thing_type = self._store.newList
             ch2 = argstr[i + 1]
             if ch2 == "$":
@@ -1046,7 +1105,7 @@ class SinkParser:
 
         return -1
 
-    def property_list(self, argstr, i, subj):
+    def property_list(self, argstr: str, i: int, subj: Node) -> int:
         """Parse property list
         Leaves the terminating punctuation in the buffer
         """
@@ -1065,19 +1124,19 @@ class SinkParser:
                 if self.turtle:
                     self.BadSyntax(argstr, j, "Found in ':-' in Turtle mode")
                 i = j + 2
-                res = []
+                res: typing.List[Any] = []
                 j = self.node(argstr, i, res, subj)
                 if j < 0:
                     self.BadSyntax(argstr, i, "bad {} or () or [] node after :- ")
                 i = j
                 continue
             i = j
-            v = []
+            v: typing.List[Any] = []
             j = self.verb(argstr, i, v)
             if j <= 0:
                 return i  # void but valid
 
-            objs = []
+            objs: typing.List[Any] = []
             i = self.objectList(argstr, j, objs)
             if i < 0:
                 self.BadSyntax(argstr, j, "objectList expected")
@@ -1095,7 +1154,13 @@ class SinkParser:
                 return i
             i += 1  # skip semicolon and continue
 
-    def commaSeparatedList(self, argstr, j, res, what):
+    def commaSeparatedList(
+        self,
+        argstr: str,
+        j: int,
+        res: MutableSequence[Any],
+        what: Callable[[str, int, MutableSequence[Any]], int],
+    ) -> int:
         """return value: -1 bad syntax; >1 new position in argstr
         res has things found appended
         """
@@ -1121,7 +1186,7 @@ class SinkParser:
             if i < 0:
                 self.BadSyntax(argstr, i, "bad list content")
 
-    def objectList(self, argstr, i, res):
+    def objectList(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         i = self.object(argstr, i, res)
         if i < 0:
             return -1
@@ -1135,7 +1200,7 @@ class SinkParser:
             if i < 0:
                 return i
 
-    def checkDot(self, argstr, i):
+    def checkDot(self, argstr: str, i: int) -> int:
         j = self.skipSpace(argstr, i)
         if j < 0:
             return j  # eof
@@ -1148,14 +1213,14 @@ class SinkParser:
             return j
         self.BadSyntax(argstr, j, "expected '.' or '}' or ']' at end of statement")
 
-    def uri_ref2(self, argstr, i, res):
+    def uri_ref2(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         """Generate uri from n3 representation.
 
         Note that the RDF convention of directly concatenating
         NS and local name is now used though I prefer inserting a '#'
         to make the namesapces look more like what XML folks expect.
         """
-        qn = []
+        qn: typing.List[Any] = []
         j = self.qname(argstr, i, qn)
         if j >= 0:
             pfx, ln = qn[0]
@@ -1182,7 +1247,7 @@ class SinkParser:
             return -1
 
         if argstr[i] == "?":
-            v = []
+            v: typing.List[Any] = []
             j = self.variable(argstr, i, v)
             if j > 0:  # Forget variables as a class, only in context.
                 res.append(v[0])
@@ -1224,7 +1289,7 @@ class SinkParser:
         else:
             return -1
 
-    def skipSpace(self, argstr, i):
+    def skipSpace(self, argstr: str, i: int) -> int:
         """Skip white space, newlines and comments.
         return -1 if EOF, else position of first non-ws character"""
 
@@ -1253,7 +1318,7 @@ class SinkParser:
         m = eof.match(argstr, i)
         return i if m is None else -1
 
-    def variable(self, argstr, i, res):
+    def variable(self, argstr: str, i: int, res) -> int:
         """?abc -> variable(:abc)"""
 
         j = self.skipSpace(argstr, i)
@@ -1270,9 +1335,10 @@ class SinkParser:
         while i < len_argstr and argstr[i] not in _notKeywordsChars:
             i += 1
         if self._parentContext is None:
-            varURI = self._store.newSymbol(self._baseURI + "#" + argstr[j:i])
+            varURI = self._store.newSymbol(self._baseURI + "#" + argstr[j:i])  # type: ignore[operator]
             if varURI not in self._variables:
-                self._variables[varURI] = self._context.newUniversal(
+                # type error: Item "None" of "Optional[Formula]" has no attribute "newUniversal"
+                self._variables[varURI] = self._context.newUniversal(  # type: ignore[union-attr]
                     varURI, why=self._reason2
                 )
             res.append(self._variables[varURI])
@@ -1281,7 +1347,7 @@ class SinkParser:
             # self.BadSyntax(argstr, j,
             #     "Can't use ?xxx syntax for variable in outermost level: %s"
             #     % argstr[j-1:i])
-        varURI = self._store.newSymbol(self._baseURI + "#" + argstr[j:i])
+        varURI = self._store.newSymbol(self._baseURI + "#" + argstr[j:i])  # type: ignore[operator]
         if varURI not in self._parentVariables:
             self._parentVariables[varURI] = self._parentContext.newUniversal(
                 varURI, why=self._reason2
@@ -1289,7 +1355,7 @@ class SinkParser:
         res.append(self._parentVariables[varURI])
         return i
 
-    def bareWord(self, argstr, i, res):
+    def bareWord(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         """abc -> :abc"""
         j = self.skipSpace(argstr, i)
         if j < 0:
@@ -1304,7 +1370,12 @@ class SinkParser:
         res.append(argstr[j:i])
         return i
 
-    def qname(self, argstr, i, res):
+    def qname(
+        self,
+        argstr: str,
+        i: int,
+        res: MutableSequence[Union[Identifier, Tuple[str, str]]],
+    ) -> int:
         """
         xyz:def -> ('xyz', 'def')
         If not in keywords and keywordsSet: def -> ('', 'def')
@@ -1362,7 +1433,7 @@ class SinkParser:
                         if c not in escapeChars:
                             raise BadSyntax(
                                 self._thisDoc,
-                                self.line,
+                                self.lines,
                                 argstr,
                                 i,
                                 "illegal escape " + c,
@@ -1374,7 +1445,7 @@ class SinkParser:
                         ):
                             raise BadSyntax(
                                 self._thisDoc,
-                                self.line,
+                                self.lines,
                                 argstr,
                                 i,
                                 "illegal hex escape " + c,
@@ -1386,7 +1457,7 @@ class SinkParser:
 
             if lastslash:
                 raise BadSyntax(
-                    self._thisDoc, self.line, argstr, i, "qname cannot end with \\"
+                    self._thisDoc, self.lines, argstr, i, "qname cannot end with \\"
                 )
 
             if argstr[i - 1] == ".":
@@ -1407,7 +1478,12 @@ class SinkParser:
                 return i
             return -1
 
-    def object(self, argstr, i, res):
+    def object(
+        self,
+        argstr: str,
+        i: int,
+        res: MutableSequence[Any],
+    ) -> int:
         j = self.subject(argstr, i, res)
         if j >= 0:
             return j
@@ -1430,12 +1506,12 @@ class SinkParser:
 
                 j, s = self.strconst(argstr, i, delim)
 
-                res.append(self._store.newLiteral(s))
+                res.append(self._store.newLiteral(s))  # type: ignore[call-arg] # TODO FIXME
                 return j
             else:
                 return -1
 
-    def nodeOrLiteral(self, argstr, i, res):
+    def nodeOrLiteral(self, argstr: str, i: int, res: MutableSequence[Any]) -> int:
         j = self.node(argstr, i, res)
         startline = self.lines  # Remember where for error messages
         if j >= 0:
@@ -1495,7 +1571,7 @@ class SinkParser:
                     lang = argstr[j + 1 : i]
                     j = i
                 if argstr[j : j + 2] == "^^":
-                    res2 = []
+                    res2: typing.List[Any] = []
                     j = self.uri_ref2(argstr, j + 2, res2)  # Read datatype URI
                     dt = res2[0]
                 res.append(self._store.newLiteral(s, dt, lang))
@@ -1503,13 +1579,13 @@ class SinkParser:
             else:
                 return -1
 
-    def uriOf(self, sym):
+    def uriOf(self, sym: Union[Identifier, Tuple[str, str]]) -> str:
         if isinstance(sym, tuple):
             return sym[1]  # old system for --pipe
         # return sym.uriref()  # cwm api
         return sym
 
-    def strconst(self, argstr, i, delim):
+    def strconst(self, argstr: str, i: int, delim: str) -> Tuple[int, str]:
         """parse an N3 string constant delimited by delim.
         return index, val
         """
@@ -1620,14 +1696,22 @@ class SinkParser:
 
         self.BadSyntax(argstr, i, "unterminated string literal")
 
-    def _unicodeEscape(self, argstr, i, startline, reg, n, prefix):
+    def _unicodeEscape(
+        self,
+        argstr: str,
+        i: int,
+        startline: int,
+        reg: Pattern[str],
+        n: int,
+        prefix: str,
+    ) -> Tuple[int, str]:
         if len(argstr) < i + n:
             raise BadSyntax(
                 self._thisDoc, startline, argstr, i, "unterminated string literal(3)"
             )
         try:
             return i + n, reg.sub(unicodeExpand, "\\" + prefix + argstr[i : i + n])
-        except:
+        except Exception:
             raise BadSyntax(
                 self._thisDoc,
                 startline,
@@ -1636,13 +1720,13 @@ class SinkParser:
                 "bad string literal hex escape: " + argstr[i : i + n],
             )
 
-    def uEscape(self, argstr, i, startline):
+    def uEscape(self, argstr: str, i: int, startline: int) -> Tuple[int, str]:
         return self._unicodeEscape(argstr, i, startline, unicodeEscape4, 4, "u")
 
-    def UEscape(self, argstr, i, startline):
+    def UEscape(self, argstr: str, i: int, startline: int) -> Tuple[int, str]:
         return self._unicodeEscape(argstr, i, startline, unicodeEscape8, 8, "U")
 
-    def BadSyntax(self, argstr, i, msg):
+    def BadSyntax(self, argstr: str, i: int, msg: str) -> NoReturn:
         raise BadSyntax(self._thisDoc, self.lines, argstr, i, msg)
 
 
@@ -1650,15 +1734,15 @@ class SinkParser:
 # [  is  operator:plus  of (  \1  \2 ) ]
 
 
-class BadSyntax(SyntaxError):
-    def __init__(self, uri, lines, argstr, i, why):
+class BadSyntax(SyntaxError):  # noqa: N818
+    def __init__(self, uri: str, lines: int, argstr: str, i: int, why: str):
         self._str = argstr.encode("utf-8")  # Better go back to strings for errors
         self._i = i
         self._why = why
         self.lines = lines
         self._uri = uri
 
-    def __str__(self):
+    def __str__(self) -> str:
         argstr = self._str
         i = self._i
         st = 0
@@ -1672,8 +1756,9 @@ class BadSyntax(SyntaxError):
         else:
             post = ""
 
+        # type error: On Python 3 formatting "b'abc'" with "%s" produces "b'abc'", not "abc"; use "%r" if this is desired behavior
         return 'at line %i of <%s>:\nBad syntax (%s) at ^ in:\n"%s%s^%s%s"' % (
-            self.lines + 1,
+            self.lines + 1,  # type: ignore[str-bytes-safe]
             self._uri,
             self._why,
             pre,
@@ -1683,31 +1768,33 @@ class BadSyntax(SyntaxError):
         )
 
     @property
-    def message(self):
+    def message(self) -> str:
         return str(self)
 
 
 ###############################################################################
-class Formula(object):
+class Formula:
     number = 0
 
-    def __init__(self, parent):
+    def __init__(self, parent: Graph):
         self.uuid = uuid4().hex
         self.counter = 0
         Formula.number += 1
         self.number = Formula.number
-        self.existentials = {}
-        self.universals = {}
+        self.existentials: Dict[str, BNode] = {}
+        self.universals: Dict[str, BNode] = {}
 
         self.quotedgraph = QuotedGraph(store=parent.store, identifier=self.id())
 
-    def __str__(self):
+    def __str__(self) -> str:
         return "_:Formula%s" % self.number
 
-    def id(self):
+    def id(self) -> BNode:
         return BNode("_:Formula%s" % self.number)
 
-    def newBlankNode(self, uri=None, why=None):
+    def newBlankNode(
+        self, uri: Optional[str] = None, why: Optional[Any] = None
+    ) -> BNode:
         if uri is None:
             self.counter += 1
             bn = BNode("f%sb%s" % (self.uuid, self.counter))
@@ -1715,28 +1802,27 @@ class Formula(object):
             bn = BNode(uri.split("#").pop().replace("_", "b"))
         return bn
 
-    def newUniversal(self, uri, why=None):
+    def newUniversal(self, uri: str, why: Optional[Any] = None) -> Variable:
         return Variable(uri.split("#").pop())
 
-    def declareExistential(self, x):
+    def declareExistential(self, x: str) -> None:
         self.existentials[x] = self.newBlankNode()
 
-    def close(self):
-
+    def close(self) -> QuotedGraph:
         return self.quotedgraph
 
 
 r_hibyte = re.compile(r"([\x80-\xff])")
 
 
-class RDFSink(object):
-    def __init__(self, graph):
-        self.rootFormula = None
+class RDFSink:
+    def __init__(self, graph: Graph):
+        self.rootFormula: Optional[Formula] = None
         self.uuid = uuid4().hex
         self.counter = 0
         self.graph = graph
 
-    def newFormula(self):
+    def newFormula(self) -> Formula:
         fa = getattr(self.graph.store, "formula_aware", False)
         if not fa:
             raise ParserError(
@@ -1745,13 +1831,18 @@ class RDFSink(object):
         f = Formula(self.graph)
         return f
 
-    def newGraph(self, identifier):
+    def newGraph(self, identifier: Identifier) -> Graph:
         return Graph(self.graph.store, identifier)
 
-    def newSymbol(self, *args):
+    def newSymbol(self, *args: str) -> URIRef:
         return URIRef(args[0])
 
-    def newBlankNode(self, arg=None, uri=None, why=None):
+    def newBlankNode(
+        self,
+        arg: Optional[Union[Formula, Graph, Any]] = None,
+        uri: Optional[str] = None,
+        why: Optional[Callable[[], None]] = None,
+    ) -> BNode:
         if isinstance(arg, Formula):
             return arg.newBlankNode(uri)
         elif isinstance(arg, Graph) or arg is None:
@@ -1761,13 +1852,13 @@ class RDFSink(object):
             bn = BNode(str(arg[0]).split("#").pop().replace("_", "b"))
         return bn
 
-    def newLiteral(self, s, dt, lang):
+    def newLiteral(self, s: str, dt: Optional[URIRef], lang: Optional[str]) -> Literal:
         if dt:
             return Literal(s, datatype=dt)
         else:
             return Literal(s, lang=lang)
 
-    def newList(self, n, f):
+    def newList(self, n: typing.List[Any], f: Optional[Formula]) -> IdentifiedNode:
         nil = self.newSymbol("http://www.w3.org/1999/02/22-rdf-syntax-ns#nil")
         if not n:
             return nil
@@ -1785,21 +1876,26 @@ class RDFSink(object):
         self.makeStatement((f, rest, a, nil))
         return af
 
-    def newSet(self, *args):
+    def newSet(self, *args: _AnyT) -> Set[_AnyT]:
         return set(args)
 
-    def setDefaultNamespace(self, *args):
+    def setDefaultNamespace(self, *args: bytes) -> str:
         return ":".join(repr(n) for n in args)
 
-    def makeStatement(self, quadruple, why=None):
+    def makeStatement(
+        self,
+        quadruple: Tuple[Optional[Union[Formula, Graph]], Node, Node, Node],
+        why: Optional[Any] = None,
+    ) -> None:
         f, p, s, o = quadruple
 
         if hasattr(p, "formula"):
             raise ParserError("Formula used as predicate")
 
-        s = self.normalise(f, s)
-        p = self.normalise(f, p)
-        o = self.normalise(f, o)
+        # type error: Argument 1 to "normalise" of "RDFSink" has incompatible type "Union[Formula, Graph, None]"; expected "Optional[Formula]"
+        s = self.normalise(f, s)  # type: ignore[arg-type]
+        p = self.normalise(f, p)  # type: ignore[arg-type]
+        o = self.normalise(f, o)  # type: ignore[arg-type]
 
         if f == self.rootFormula:
             # print s, p, o, '.'
@@ -1807,11 +1903,16 @@ class RDFSink(object):
         elif isinstance(f, Formula):
             f.quotedgraph.add((s, p, o))
         else:
-            f.add((s, p, o))
+            # type error: Item "None" of "Optional[Graph]" has no attribute "add"
+            f.add((s, p, o))  # type: ignore[union-attr]
 
         # return str(quadruple)
 
-    def normalise(self, f, n):
+    def normalise(
+        self,
+        f: Optional[Formula],
+        n: Union[Tuple[int, str], bool, int, Decimal, float, _AnyT],
+    ) -> Union[URIRef, Literal, BNode, _AnyT]:
         if isinstance(n, tuple):
             return URIRef(str(n[1]))
 
@@ -1836,6 +1937,8 @@ class RDFSink(object):
 
         if isinstance(f, Formula):
             if n in f.existentials:
+                if TYPE_CHECKING:
+                    assert isinstance(n, URIRef)
                 return f.existentials[n]
 
         # if isinstance(n, Var):
@@ -1843,19 +1946,19 @@ class RDFSink(object):
         #       return f.universals[n]
         #    f.universals[n] = f.newBlankNode()
         #    return f.universals[n]
+        # type error: Incompatible return value type (got "Union[int, _AnyT]", expected "Union[URIRef, Literal, BNode, _AnyT]")  [return-value]
+        return n  # type: ignore[return-value]
 
-        return n
-
-    def intern(self, something):
+    def intern(self, something: _AnyT) -> _AnyT:
         return something
 
-    def bind(self, pfx, uri):
+    def bind(self, pfx, uri) -> None:
         pass  # print pfx, ':', uri
 
-    def startDoc(self, formula):
+    def startDoc(self, formula: Optional[Formula]) -> None:
         self.rootFormula = formula
 
-    def endDoc(self, formula):
+    def endDoc(self, formula: Optional[Formula]) -> None:
         pass
 
 
@@ -1865,7 +1968,7 @@ class RDFSink(object):
 #
 
 
-def hexify(ustr):
+def hexify(ustr: str) -> bytes:
     """Use URL encoding to return an ASCII string
     corresponding to the given UTF8 string
 
@@ -1885,7 +1988,6 @@ def hexify(ustr):
 
 
 class TurtleParser(Parser):
-
     """
     An RDFLib parser for Turtle
 
@@ -1895,8 +1997,13 @@ class TurtleParser(Parser):
     def __init__(self):
         pass
 
-    def parse(self, source, graph, encoding="utf-8", turtle=True):
-
+    def parse(
+        self,
+        source: InputSource,
+        graph: Graph,
+        encoding: Optional[str] = "utf-8",
+        turtle: bool = True,
+    ) -> None:
         if encoding not in [None, "utf-8"]:
             raise ParserError(
                 "N3/Turtle files are always utf-8 encoded, I was passed: %s" % encoding
@@ -1917,7 +2024,6 @@ class TurtleParser(Parser):
 
 
 class N3Parser(TurtleParser):
-
     """
     An RDFLib parser for Notation3
 
@@ -1928,7 +2034,10 @@ class N3Parser(TurtleParser):
     def __init__(self):
         pass
 
-    def parse(self, source, graph, encoding="utf-8"):
+    # type error: Signature of "parse" incompatible with supertype "TurtleParser"
+    def parse(  # type: ignore[override]
+        self, source: InputSource, graph: Graph, encoding: Optional[str] = "utf-8"
+    ) -> None:
         # we're currently being handed a Graph, not a ConjunctiveGraph
         # context-aware is this implied by formula_aware
         ca = getattr(graph.store, "context_aware", False)
@@ -1938,41 +2047,10 @@ class N3Parser(TurtleParser):
         elif not fa:
             raise ParserError("Cannot parse N3 into non-formula-aware store.")
 
-        conj_graph = ConjunctiveGraph(store=graph.store)
+        conj_graph = Dataset(store=graph.store)
         conj_graph.default_context = graph  # TODO: CG __init__ should have a
         # default_context arg
         # TODO: update N3Processor so that it can use conj_graph as the sink
         conj_graph.namespace_manager = graph.namespace_manager
 
         TurtleParser.parse(self, source, conj_graph, encoding, turtle=False)
-
-
-def _test():  # pragma: no cover
-    import doctest
-
-    doctest.testmod()
-
-
-def main():  # pragma: no cover
-    g = ConjunctiveGraph()
-
-    sink = RDFSink(g)
-    base_uri = "file://" + os.path.join(os.getcwd(), sys.argv[1])
-
-    p = SinkParser(sink, baseURI=base_uri)
-    p._bindings[""] = p._baseURI + "#"
-    p.startDoc()
-
-    f = open(sys.argv[1], "rb")
-    rdbytes = f.read()
-    f.close()
-
-    p.feed(rdbytes)
-    p.endDoc()
-    for t in g.quads((None, None, None)):
-
-        print(t)
-
-
-if __name__ == "__main__":
-    main()
