@@ -1,24 +1,24 @@
 #!/usr/bin/env python
-# -*- coding: UTF-8 -*-
+
 """
 A commandline tool for parsing RDF in different formats and serializing the
 resulting graph to a chosen format.
 """
+from __future__ import annotations
 
+import logging
 import sys
 from optparse import OptionParser
-import logging
+from typing import BinaryIO, Optional
 
 import rdflib
 from rdflib import plugin
-from rdflib.store import Store
 from rdflib.graph import ConjunctiveGraph
 from rdflib.parser import Parser
 from rdflib.serializer import Serializer
-
+from rdflib.store import Store
 from rdflib.util import guess_format
 
-
 DEFAULT_INPUT_FORMAT = "xml"
 DEFAULT_OUTPUT_FORMAT = "n3"
 
@@ -33,7 +33,6 @@ def parse_and_serialize(
     store_conn="",
     store_type=None,
 ):
-
     if store_type:
         store = plugin.get(store_type, Store)()
         store.open(store_conn)
@@ -192,7 +191,7 @@ def main():
             pfx, uri = ns_kw.split("=")
             ns_bindings[pfx] = uri
 
-    outfile = sys.stdout.buffer
+    outfile: Optional[BinaryIO] = sys.stdout.buffer
 
     if opts.no_out:
         outfile = None
