@@ -9,7 +9,7 @@ from pytest import MonkeyPatch
 import rdflib.plugins.sparql
 import rdflib.plugins.sparql.operators
 import rdflib.plugins.sparql.parser
-from rdflib import BNode, ConjunctiveGraph, Graph, Literal, URIRef
+from rdflib import BNode, Dataset, Graph, Literal, URIRef
 from rdflib.compare import isomorphic
 from rdflib.namespace import RDF, RDFS, Namespace
 from rdflib.plugins.sparql import prepareQuery, sparql
@@ -182,11 +182,11 @@ def test_bindings():
 
 
 def test_named_filter_graph_query():
-    g = ConjunctiveGraph()
-    g.namespace_manager.bind("rdf", RDF)
-    g.namespace_manager.bind("rdfs", RDFS)
+    g = Dataset(default_union=True)
+    g.bind("rdf", RDF)
+    g.bind("rdfs", RDFS)
     ex = Namespace("https://ex.com/")
-    g.namespace_manager.bind("ex", ex)
+    g.bind("ex", ex)
     g.get_context(ex.g1).parse(
         format="turtle",
         data=f"""
